import numpy as np


class Obstacle:
    def __init__(
        self,
        is_agent,
        velocity,
        theta,
        initial_position,
        simulation_time,
        number_of_timesteps,
        radius=0.5,
    ):
        self.is_agent = is_agent
        self.velocity = velocity
        self.initial_position = np.array([initial_position.x, initial_position.y])
        self.theta = theta
        self.radius = radius

        # Creates an obstacle starting at initial position and moving at velocity in theta direction
        t = np.linspace(0, simulation_time, number_of_timesteps)
        theta = self.theta * np.ones(np.shape(t))
        vx = self.velocity * np.cos(theta)
        vy = self.velocity * np.sin(theta)
        v = np.stack([vx, vy])
        p0 = self.initial_position.reshape((2, 1))
        p = p0 + np.cumsum(v, axis=1) * (simulation_time / number_of_timesteps)
        self.state = np.concatenate((p, v)).reshape(4, number_of_timesteps, 1)

    def __str__(self):
        return str(self.state)



"""
Decentralized planning using velocity obstacles 
author: Erwin Lejeune (erwin.lejeune15@gmail.com)
"""

from .obstacle import Obstacle
from .velocity_agent import VelocityAgent
import numpy as np
import matplotlib as mpl
from random import random
import os

if "DISPLAY" not in os.environ:
    mpl.use("agg")
else:
    mpl.use("TkAgg")
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from matplotlib.patches import Circle
import coloredlogs
import logging


class MultiAgentVelocityObstacle:
    def __init__(self, simulation_time=8.0, timestep=0.1, log_level="WARNING"):
        self.simulation_time = simulation_time
        self.timestep = timestep
        self.number_of_timesteps = int(simulation_time / timestep)

        # Agents
        self.agents = dict()
        self.global_state_history = dict()

        # Obstacles
        self.obstacles_objects = []

        # Flags
        self.simulation_complete = False
        self.__init_logger(log_level)
        coloredlogs.install(level=log_level)

    def register_agent(self, id, start, goal, radius=0.5, vmax=2, vmin=0.2):
        if id in self.agents:
            logging.warning("Failed to register agent: ID has already been registered.")
            return
        agent = VelocityAgent(
            id, start, goal, self.number_of_timesteps, self.timestep, radius, vmax, vmin
        )
        self.agents[id] = agent

    def register_obstacle(self, velocity, theta, initial_position):
        obstacle = Obstacle(
            False,
            velocity,
            theta,
            initial_position,
            self.simulation_time,
            self.number_of_timesteps,
        )
        self.obstacles_objects.append(obstacle)
        if not hasattr(self, "obstacles"):
            self.obstacles = obstacle.state
        else:
            self.obstacles = np.dstack((self.obstacles, obstacle.state))

    def run_simulation(self):
        try:
            obstacles = self.obstacles
        except:
            obstacles = []
        for i in range(self.number_of_timesteps):
            other_agents = []
            for key, agent in self.agents.items():
                state_history, vel, state = agent.simulate_step(
                    i, obstacles, other_agents
                )
                agent_as_obstacle = self.__agent_to_obstacle(vel, state)
                other_agents.append(agent_as_obstacle)
                self.global_state_history[key] = state_history
        self.simulation_complete = True

    def visualize(self, saved_file, map_length, map_height):
        if not self.simulation_complete:
            logging.warning("Failed to visualize: simulation must be complete")
            return

        self.__plot(saved_file, map_length, map_height)

    def __agent_to_obstacle(self, velocity, pos):
        return np.concatenate((pos, velocity))

    def __init_logger(self, log_level):
        import os

        if not os.path.exists("logs"):
            os.makedirs("logs")
        logging.basicConfig(
            filename="logs/main.log",
            format="%(levelname)s:%(message)s",
            filemode="w",
            level=log_level,
        )

    def __plot(self, saved_file, map_length, map_height):
        fig = plt.figure()
        ax = fig.add_subplot(
            111, autoscale_on=False, xlim=(0, map_length), ylim=(0, map_height)
        )
        ax.set_aspect("equal")
        ax.grid()

        lines = []
        agents_list = []
        for key, agent in self.agents.items():
            rgb = [random(), random(), random()]
            agents_list.append(
                Circle(
                    (
                        self.global_state_history[key][0, 0],
                        self.global_state_history[key][0, 1],
                    ),
                    agent.radius,
                    facecolor=rgb,
                    edgecolor="black",
                )
            )
            (line,) = ax.plot([], [], c=rgb)
            lines.append(line)
        obstacle_list = []
        try:
            for obstacle in range(np.shape(self.obstacles)[2]):
                obstacle_list.append(
                    Circle(
                        (0, 0),
                        self.obstacles_objects[obstacle].radius,
                        facecolor="red",
                        edgecolor="black",
                    )
                )
        except:
            pass

        def init():
            for agent in agents_list:
                ax.add_patch(agent)
                line.set_data([], [])
                lines.append(line)
            for obstacle in obstacle_list:
                ax.add_patch(obstacle)

            return agents_list + lines + obstacle_list

        def animate(i):
            k = 0
            for key, _ in self.agents.items():
                agents_list[k].center = (
                    self.global_state_history[key][0, i],
                    self.global_state_history[key][1, i],
                )
                lines[k].set_data(
                    self.global_state_history[key][0, :i],
                    self.global_state_history[key][1, :i],
                )
                k += 1
            for j in range(len(obstacle_list)):
                obstacle_list[j].center = (
                    self.obstacles[0, i, j],
                    self.obstacles[1, i, j],
                )
            return agents_list + lines + obstacle_list

        init()
        step = self.simulation_time / self.number_of_timesteps
        for i in range(self.number_of_timesteps):
            animate(i)
            plt.pause(step)

        # Save animation
        if not saved_file:
            return

        ani = animation.FuncAnimation(
            fig,
            animate,
            np.arange(1, self.number_of_timesteps),
            interval=200,
            blit=True,
            init_func=init,
        )

        ani.save(saved_file + str(".gif"), writer="ffmpeg", fps=10)



import numpy as np


class VelocityAgent:
    def __init__(
        self,
        id,
        start,
        goal,
        number_of_timesteps,
        timestep,
        radius=0.5,
        vmax=2,
        vmin=0.2,
    ):
        # Agent Initialization
        self.id = id
        self.start = np.array([start.x, start.y])
        self.goal = np.array([goal.x, goal.y])

        # MAPF Velocity Obstacles Consts
        self.number_of_timesteps = number_of_timesteps
        self.timestep = timestep

        # Agent Consts
        self.radius = radius
        self.vmax = vmax
        self.vmin = vmin

        # Current State
        self.current_state = self.start
        self.state_history = np.empty((4, self.number_of_timesteps))

    def simulate_step(self, step, obstacles, other_agents):
        # Predict Obstacles and Agents Positions in the Future
        v_desired = self.__compute_desired_velocity(self.current_state[:2], self.goal)
        control_vel = self.__compute_velocity(
            self.current_state, obstacles, step, v_desired, other_agents
        )
        self.current_state = self.__update_state(self.current_state, control_vel)
        self.state_history[:4, step] = self.current_state
        return self.state_history, self.current_state, control_vel

    def __compute_desired_velocity(self, current_pos, goal_pos):
        disp_vec = (goal_pos - current_pos)[:2]
        norm = np.linalg.norm(disp_vec)
        if norm < self.radius / 5:
            return np.zeros(2)
        disp_vec = disp_vec / norm
        np.shape(disp_vec)
        desired_vel = self.vmax * disp_vec
        return desired_vel

    def __compute_velocity(self, state, obstacles, step, v_desired, other_agents):
        try:
            obstacles = obstacles[:, step, :]
            pos_agent = state[:2]
            number_of_obstacles = np.shape(obstacles)[1]

        except:
            pos_agent = state[:2]
            number_of_obstacles = 0

        a = np.empty(((number_of_obstacles + len(other_agents)) * 2, 2))
        b = np.empty(((number_of_obstacles + len(other_agents)) * 2))
        # Handle none agents Obstacles
        for i in range(number_of_obstacles):
            obstacle = obstacles[:, i]
            pos_obs = obstacle[:2]
            vel_obs = obstacle[2:]
            dispBA = pos_agent - pos_obs
            distBA = np.linalg.norm(dispBA)
            thetaBA = np.arctan2(dispBA[1], dispBA[0])

            if 2.2 * self.radius > distBA:
                distBA = 2.2 * self.radius

            phi_obst = np.arcsin(2.2 * self.radius / distBA)
            phi_left = thetaBA + phi_obst
            phi_right = thetaBA - phi_obst

            # VO
            translation = vel_obs
            a_temp, b_temp = self.__create_constraints(translation, phi_left, "left")
            a[i * 2, :] = a_temp
            b[i * 2] = b_temp
            a_temp, b_temp = self.__create_constraints(translation, phi_right, "right")
            a[i * 2 + 1, :] = a_temp
            b[i * 2 + 1] = b_temp

        # Handle other agents
        k = number_of_obstacles
        for obs_agent in other_agents:
            obs_agent_position = obs_agent[:2]
            obs_agent_velocity = obs_agent[2:]
            dispBA = pos_agent - obs_agent_position
            distBA = np.linalg.norm(dispBA)
            thetaBA = np.arctan2(dispBA[1], dispBA[0])

            if 2.2 * self.radius > distBA:
                distBA = 2.2 * self.radius

            phi_obst = np.arcsin(2.2 * self.radius / distBA)
            phi_left = thetaBA + phi_obst
            phi_right = thetaBA - phi_obst

            # VO
            translation = obs_agent_velocity
            a_temp, b_temp = self.__create_constraints(translation, phi_left, "left")
            a[k * 2, :] = a_temp
            b[k * 2] = b_temp
            a_temp, b_temp = self.__create_constraints(translation, phi_right, "right")
            a[k * 2 + 1, :] = a_temp
            b[k * 2 + 1] = b_temp
            k += 1

        # Create search-space
        th = np.linspace(0, 2 * np.pi, 20)
        vel = np.linspace(0, self.vmax, 5)
        vv, thth = np.meshgrid(vel, th)
        vx_sample = (vv * np.cos(thth)).flatten()
        vy_sample = (vv * np.sin(thth)).flatten()
        v_sample = np.stack((vx_sample, vy_sample))

        try:
            v_satisfying_constraints = self.__check_constraints(v_sample, a, b)
            # Objective function
            size = np.shape(v_satisfying_constraints)[1]
            diffs = v_satisfying_constraints - (
                (v_desired).reshape(2, 1) @ np.ones(size).reshape(1, size)
            )
            norm = np.linalg.norm(diffs, axis=0)
            min_index = np.where(norm == np.amin(norm))[0][0]
            cmd_vel = v_satisfying_constraints[:, min_index]
        except:
            cmd_vel = np.array([0, 0])
        return cmd_vel

    def __check_constraints(self, v_sample, Amat, bvec):
        length = np.shape(bvec)[0]

        for i in range(int(length / 2)):
            v_sample = self.__check_inside(
                v_sample, Amat[2 * i : 2 * i + 2, :], bvec[2 * i : 2 * i + 2]
            )

        return v_sample

    def __check_inside(self, v, Amat, bvec):
        v_out = []
        for i in range(np.shape(v)[1]):
            if not ((Amat @ v[:, i] < bvec).all()):
                v_out.append(v[:, i])
        return np.array(v_out).T

    def __create_constraints(self, translation, angle, side):
        # create line
        origin = np.array([0, 0, 1])
        point = np.array([np.cos(angle), np.sin(angle)])
        line = np.cross(origin, point)
        line = self.__translate_line(line, translation)

        if side == "left":
            line *= -1

        A = line[:2]
        b = -line[2]

        return A, b

    def __translate_line(self, line, translation):
        matrix = np.eye(3)
        matrix[2, :2] = -translation[:2]
        return matrix @ line

    def __update_state(self, x, v):
        new_state = np.empty((4))
        new_state[:2] = x[:2] + v * self.timestep
        new_state[-2:] = v
        return new_state



import numpy as np
from scipy.optimize import minimize, Bounds
import time


class NMPCAgent:
    def __init__(
        self,
        id,
        start,
        goal,
        number_of_timesteps,
        nmpc_timestep,
        timestep,
        qc=5.0,
        kappa=4.0,
        radius=0.5,
        vmax=2,
        vmin=0.2,
        horizon_length=4,
    ):
        # Initializations
        self.id = id
        self.start = np.array([start.x, start.y])
        self.goal = np.array([goal.x, goal.y])

        # Consts from NMPC Class
        self.number_of_timesteps = number_of_timesteps
        self.nmpc_timestep = nmpc_timestep
        self.timestep = timestep

        # Agent Constraints
        self.horizon_length = horizon_length
        self.vmax = vmax
        self.vmin = vmin
        self.qc = qc
        self.kappa = kappa
        self.upper_bound = [(1 / np.sqrt(2)) * self.vmax] * self.horizon_length * 2
        self.lower_bound = [-(1 / np.sqrt(2)) * self.vmax] * self.horizon_length * 2
        self.radius = radius

        # Current State
        self.current_state = self.start
        self.state_history = np.empty((4, self.number_of_timesteps))

    def simulate_step(self, step, obstacles, other_agents):
        # Predict Obstacles and Agents Positions in the Future
        obstacle_prediction = self.__predict_obstacle_positions(
            obstacles, step, other_agents
        )
        xref = self.__compute_xref()
        vel, _ = self.__compute_velocity(self.current_state, obstacle_prediction, xref)
        self.current_state = self.__update_state(self.current_state, vel)
        self.state_history[:2, step] = self.current_state
        return self.state_history, vel, self.current_state

    def __compute_velocity(self, robot_state, obstacle_predictions, xref):
        u0 = np.random.rand(2 * self.horizon_length)

        def cost_fn(u):
            return self.__total_cost(u, robot_state, obstacle_predictions, xref)

        bounds = Bounds(self.lower_bound, self.upper_bound)

        res = minimize(cost_fn, u0, method="SLSQP", bounds=bounds)
        velocity = res.x[:2]
        return velocity, res.x

    def __compute_xref(self):
        dir_vec = self.goal - self.current_state
        norm = np.linalg.norm(dir_vec)
        if norm < 0.1:
            new_goal = self.current_state
        else:
            dir_vec = dir_vec / norm
            new_goal = (
                self.current_state
                + dir_vec * self.vmax * self.nmpc_timestep * self.horizon_length
            )
        return np.linspace(self.current_state, new_goal, self.horizon_length).reshape(
            (2 * self.horizon_length)
        )

    def __total_cost(self, u, robot_state, obstacle_predictions, xref):
        x_robot = self.__update_state(robot_state, u)
        c1 = self.__tracking_cost(x_robot, xref)
        c2 = self.__total_collision_cost(x_robot, obstacle_predictions)
        total = c1 + c2
        return total

    def __tracking_cost(self, x, xref):
        return np.linalg.norm(x - xref)

    def __total_collision_cost(self, robot, obstacles):
        total_cost = 0
        for i in range(self.horizon_length):
            for j in range(len(obstacles)):
                obstacle = obstacles[j]
                rob = robot[2 * i : 2 * i + 2]
                obs = obstacle[2 * i : 2 * i + 2]
                total_cost += self.__collision_cost(rob, obs)
        return total_cost

    def __collision_cost(self, x0, x1):
        """
        Cost of collision between two robot_state
        """
        d = np.linalg.norm(x0 - x1)
        cost = self.qc / (1 + np.exp(self.kappa * (d - 2 * self.radius)))
        return cost

    def __predict_obstacle_positions(self, obstacles, step, other_agents):
        obstacle_predictions = []
        try:
            obstacles[:, step, :]
            for i in range(np.shape(obstacles)[1]):
                obstacle = obstacles[:, i]
                obstacle_position = obstacle[:2]
                obstacle_vel = obstacle[2:]
                u = np.vstack([np.eye(2)] * self.horizon_length) @ obstacle_vel
                obstacle_prediction = self.__update_state(obstacle_position, u)
                obstacle_predictions.append(obstacle_prediction)
        except:
            pass

        for agent in other_agents:
            agent_position = agent[:2]
            agent_vel = agent[2:]
            u = np.vstack([np.eye(2)] * self.horizon_length) @ agent_vel
            obstacle_agent_prediction = self.__update_state(agent_position, u)
            obstacle_predictions.append(obstacle_agent_prediction)

        return obstacle_predictions

    def __update_state(self, x0, u):
        """
        Computes the states of the system after applying a sequence of control signals u on
        initial state x0
        """
        N = int(len(u) / 2)
        lower_triangular_ones_matrix = np.tril(np.ones((N, N)))
        kron = np.kron(lower_triangular_ones_matrix, np.eye(2))

        new_state = np.vstack([np.eye(2)] * int(N)) @ x0 + kron @ u * self.nmpc_timestep

        return new_state

    def __hash__(self):
        h = str(self.id) + str(self.start) + str(self.goal) + str(self.radius)
        return h

    def __eq__(self, other):
        return self.id == other.id

    def __str__(self):
        return "=======AGENT=======\nID = %s\nSTART_POSITION = %s\nGOAL_POSITION = %s\nRADIUS = %s\nVMIN = %s\nVMAX = %s\nHORIZON_LENGTH = %s\nQC = %s\nKappa = %s\n===================" % (
            str(self.id),
            str(self.start),
            str(self.goal),
            str(self.radius),
            str(self.vmin),
            str(self.vmax),
            str(self.horizon_length),
            str(self.qc),
            str(self.kappa),
        )



import setuptools

with open("requirements.txt") as f:
    required = f.read().splitlines()

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


setuptools.setup(
    name="pymapf",  # Replace with your own username
    version="0.1.2",
    author="Erwin Lejeune",
    author_email="erwinlejeune.pro@gmail.com",
    description="Pymapf is a library for decentralized and centralized multi-agent planning tools",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/apla-toolbox/pymapf",
    packages=setuptools.find_packages(),
    install_requires=required,
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: Unix",
        "Operating System :: MacOS",
        "Framework :: Pytest",
    ],
    python_requires=">=3.6",
)
